@@ -28,11 +28,12 @@
 //!     .collect::<VecDeque<_>>();
 //!
 //! let mut sample_index = 0;
-//! while let (initial_sample_buffer_len, Some(estimated_pitch)) = (
+//! while let (initial_sample_buffer_len, Some(output)) = (
 //!     sample_buffer.len(),
 //!     irapt.process(&mut sample_buffer),
 //! ) {
-//!     let estimated_pitch_index = sample_index + estimated_pitch.index;
+//!     let estimated_pitch = output.pitch_estimates().final_estimate();
+//!     let estimated_pitch_index = (sample_index as isize + estimated_pitch.offset) as usize;
 //!     let estimated_pitch_time = estimated_pitch_index as f64 / parameters.sample_rate;
 //!     println!("estimated pitch at {:0.3}: {}Hz with energy {}",
 //!              estimated_pitch_time, estimated_pitch.frequency, estimated_pitch.energy);
@@ -57,12 +58,12 @@ pub mod interpolate;
 #[doc(hidden)]
 pub mod polyphase_filter;
 
-use self::candidates::CandidateGenerator;
-use self::candidates::CandidateSelector;
+use self::candidates::{CandidateFrequencyIter, CandidateGenerator, CandidateSelectionStepIter, CandidateSelector};
 use self::error::InvalidParameterError;
 use self::harmonics::HarmonicParametersEstimator;
 
 use alloc::collections::VecDeque;
+use core::iter::Enumerate;
 use core::ops::RangeInclusive;
 
 /// Implementation of the IRAPT pitch estimation algorithm.
@@ -167,14 +168,36 @@ pub struct Parameters {
     pub candidate_max_jump: usize,
 }
 
-/// Estimate of the current pitch of the input.
+/// The output of [`Irapt::process`].
+///
+/// This `struct` holds the output, including estimated pitches, of the IRAPT algorithm after processing a single time step, and is created
+/// by the [`process`](Irapt::process) method on [`Irapt`]. See its documentation for more.
+pub struct Output<'a> {
+    estimated_pitches: EstimatedPitchIter<'a>,
+    more_output: bool,
+}
+
+/// An estimate of the pitch in the input at a specific sample offset.
+#[derive(Clone, Copy, Debug, PartialEq)]
 pub struct EstimatedPitch {
     /// Frequency, in Hz, of the estimated pitch.
     pub frequency: f64,
-    /// Arbitrary measure, from `0.0..=1.0`, of the energy associated with the estimated pitch.
+    /// Arbitrary measure, from `0.0..`, of the energy associated with the estimated pitch.
     pub energy:    f64,
-    /// The index within the input buffer (_before_ removal of consumed samples) at which this pitch was estimated.
-    pub index:     usize,
+    /// The offset in samples within the input buffer (_before_ removal of consumed samples) at which the pitch was estimated. This may be
+    /// negative, since estimates can be returned for samples which have already been removed from the input buffer.
+    pub offset:    isize,
+}
+
+/// An iterator over pitches estimated over time in the input, in reverse chronological order.
+///
+/// This `struct` is created by the [`pitch_estimates`](Output::pitch_estimates) method on [`Output`]. See its documentation for more.
+#[derive(Clone)]
+pub struct EstimatedPitchIter<'a> {
+    selected_candidates: Enumerate<CandidateSelectionStepIter<'a>>,
+    candidate_frequencies: CandidateFrequencyIter,
+    last_step_sample_index: usize,
+    step_len: usize,
 }
 
 impl Irapt {
@@ -214,6 +237,11 @@ impl Irapt {
         })
     }
 
+    /// Returns the `Parameters` specified during construction.
+    pub fn parameters(&self) -> &Parameters {
+        &self.parameters
+    }
+
     /// Process input from a queue of samples in a [`VecDeque`].
     ///
     /// As many samples as necessary to calculate the next pitch estimate are read from the [`VecDeque`], otherwise
@@ -242,22 +270,26 @@ impl Irapt {
     /// sample_buffer.extend(samples.by_ref().take(parameters.sample_rate as usize / 2));
     ///
     /// // Process as many samples as possible
-    /// while let Some(estimated_pitch) = irapt.process(&mut sample_buffer) {
+    /// while let Some(output) = irapt.process(&mut sample_buffer) {
+    ///     let estimated_pitch = output.pitch_estimates().final_estimate();
     ///     println!("estimated pitch: {}Hz with energy {}", estimated_pitch.frequency, estimated_pitch.energy);
     /// }
     ///
     /// // Simulate that half of a second more samples have become availoble and process them
     /// sample_buffer.extend(samples.by_ref().take(parameters.sample_rate as usize / 2));
     ///
-    /// while let Some(estimated_pitch) = irapt.process(&mut sample_buffer) {
+    /// while let Some(output) = irapt.process(&mut sample_buffer) {
+    ///     let estimated_pitch = output.pitch_estimates().final_estimate();
     ///     println!("estimated pitch: {}Hz with energy {}", estimated_pitch.frequency, estimated_pitch.energy);
     /// }
     /// ```
-    pub fn process<S: Into<f64> + Copy>(&mut self, samples: &mut VecDeque<S>) -> Option<EstimatedPitch> {
+    pub fn process<S: Into<f64> + Copy>(&mut self, samples: &mut VecDeque<S>) -> Option<Output<'_>> {
         let initial_samples_len = samples.len();
         let step_len = self.parameters.step_len();
-        while let (step_samples_len, Some(harmonics)) = (
-            self.estimator.next_step_samples_len(),
+
+        let mut processed_step_sample_index = None;
+        while let (step_sample_index, Some(harmonics)) = (
+            initial_samples_len - samples.len() + self.estimator.next_step_samples_len(),
             self.estimator.process_step(samples, step_len, self.parameters.sample_rate),
         ) {
             let mut energy = 0.0;
@@ -265,33 +297,39 @@ impl Irapt {
                 energy += harmonic.amplitude * harmonic.amplitude;
             });
 
-            let candidates = self.candidate_generator.process_step(harmonics, self.parameters.sample_rate);
+            self.candidate_generator.process_step_harmonics(harmonics, self.parameters.sample_rate);
+            let candidates = self.candidate_generator.generate_step_candidates();
 
-            let mut min_candidate = 0.0;
-            let candidates = candidates.inspect(|candidate| {
-                min_candidate = candidate.min(min_candidate);
-            });
-
-            if let Some(best_candidate_index) = self.candidate_selector.process_step(
+            self.candidate_selector.process_step(
                 candidates,
+                energy,
                 self.parameters.candidate_selection_window_len(),
                 self.parameters.candidate_max_jump,
                 self.parameters.candidate_step_decay,
-            ) {
-                let candidate_frequency = self
-                    .candidate_generator
-                    .candidate_frequencies(self.parameters.sample_rate)
-                    .nth(best_candidate_index)
-                    .unwrap_or_else(|| panic!("candidate index out of bounds"));
-                let removed_samples_count = initial_samples_len - samples.len();
-                return Some(EstimatedPitch {
-                    frequency: candidate_frequency,
-                    energy:    (-min_candidate / energy.max(1e-4)),
-                    index:     step_samples_len + removed_samples_count,
-                });
+            );
+
+            if self.candidate_selector.initialized(self.parameters.candidate_selection_window_len()) {
+                processed_step_sample_index = Some(step_sample_index);
+                break;
             }
         }
-        None
+
+        let last_step_sample_index = processed_step_sample_index?;
+        let more_output = samples.len() >= self.estimator.next_step_samples_len();
+
+        let selected_candidates = self.candidate_selector.best_candidate_steps(
+            self.parameters.candidate_selection_window_len(),
+            self.parameters.candidate_max_jump,
+        );
+        Some(Output {
+            estimated_pitches: EstimatedPitchIter {
+                selected_candidates: selected_candidates?.enumerate(),
+                candidate_frequencies: self.candidate_generator.candidate_frequencies(self.parameters.sample_rate),
+                last_step_sample_index,
+                step_len: self.parameters.step_len(),
+            },
+            more_output,
+        })
     }
 
     /// Resets all internal state associated with the sliding analysis window.
@@ -317,7 +355,8 @@ impl Irapt {
     /// let mut sample_buffer = VecDeque::new();
     /// sample_buffer.extend(samples.by_ref().take(parameters.sample_rate as usize / 2));
     ///
-    /// while let Some(estimated_pitch) = irapt.process(&mut sample_buffer) {
+    /// while let Some(output) = irapt.process(&mut sample_buffer) {
+    ///     let estimated_pitch = output.pitch_estimates().final_estimate();
     ///     println!("estimated pitch: {}Hz with energy {}", estimated_pitch.frequency, estimated_pitch.energy);
     /// }
     ///
@@ -329,12 +368,14 @@ impl Irapt {
     /// sample_buffer.clear();
     /// sample_buffer.extend(more_samples.skip(parameters.sample_rate as usize * 19 / 2));
     ///
-    /// while let Some(estimated_pitch) = irapt.process(&mut sample_buffer) {
+    /// while let Some(output) = irapt.process(&mut sample_buffer) {
+    ///     let estimated_pitch = output.pitch_estimates().final_estimate();
     ///     println!("estimated pitch: {}Hz with energy {}", estimated_pitch.frequency, estimated_pitch.energy);
     /// }
     /// ```
     pub fn reset(&mut self) {
         self.estimator.reset();
+        self.candidate_selector.reset();
     }
 }
 
@@ -380,3 +421,69 @@ impl Default for Parameters {
         Self::DEFAULT
     }
 }
+
+//
+// Output impls
+//
+
+impl Output<'_> {
+    /// Returns whether further output can be produced given the input samples.
+    ///
+    /// More output can be produced by calling [`Irapt::process`].
+    pub fn more_output(&self) -> bool {
+        self.more_output
+    }
+
+    /// Returns all pitch estimates for the given input, including both those tentative and final, in reverse chronological order.
+    ///
+    /// All but the last of the yielded pitches are tentative estimates calculated up to [`candidate_selection_window_duration`] seconds in
+    /// the past. The estimates are returned in reverse chronological order. The exact sample offsets for the estimates are returned in
+    /// [`EstimatedPitch::offset`].
+    ///
+    /// The last estimate yielded is final for the given time offset. It can also be retrieved by calling [`final_estimate`] on the returned
+    /// iterator.
+    ///
+    /// [`candidate_selection_window_duration`]: Parameters::candidate_selection_window_duration
+    /// [`final_estimate`]: EstimatedPitchIter::final_estimate
+    pub fn pitch_estimates(&self) -> EstimatedPitchIter<'_> {
+        self.estimated_pitches.clone()
+    }
+}
+
+//
+// EstimatedPitchIter impls
+//
+
+impl EstimatedPitchIter<'_> {
+    /// Returns a final pitch estimate for the given input, at a time delay.
+    ///
+    /// The returned pitch is the final estimate calculated at approximately [`candidate_selection_window_duration`] seconds in the past.
+    /// The exact sample offset for the estimate is returned in [`EstimatedPitch::offset`].
+    ///
+    /// [`candidate_selection_window_duration`]: Parameters::candidate_selection_window_duration
+    pub fn final_estimate(self) -> EstimatedPitch {
+        self.last().unwrap_or_else(|| unreachable!())
+    }
+}
+
+impl Iterator for EstimatedPitchIter<'_> {
+    type Item = EstimatedPitch;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        let (step_index, candidate_selection) = self.selected_candidates.next()?;
+        let frequency = (self.candidate_frequencies.clone())
+            .nth(candidate_selection.selected_candidate_index)
+            .unwrap_or_else(|| panic!("candidate index out of bounds"));
+        Some(EstimatedPitch {
+            frequency,
+            energy: candidate_selection.energy,
+            offset: self.last_step_sample_index.wrapping_sub((1 + step_index) * self.step_len) as isize
+        })
+    }
+
+    fn size_hint(&self) -> (usize, Option<usize>) {
+        self.selected_candidates.size_hint()
+    }
+}
+
+impl ExactSizeIterator for EstimatedPitchIter<'_> {}
